"""
This addon adapts selected statements. 
DISCLAIMER: If you try to adapt CIRC statements you will have to select  at least one other statement together with the CIRC statement.
Updating a single CIRC statement does not work due to a bug in KUKA.Sim (BugTicket 963861)!

Known issues:
  - adaption of turn values does not work for unknown reasons. Probably turn is handled different for robots in KUKA.Sim/VC and KUKA KSS environment.

"""

import os
import sys
from vcCommand import *
from vcHelpers.Selection import *
from kuka import *
import vcMatrix
import kuka_sim_extensions.Libs.DS_Library.ds_sim_helpers as ds_helpers

GLOBAL_ERROR = "ADAPT STATEMENTS ERROR - "

global init, app
global scale_switch, x_switch, y_switch, z_switch, a_switch, b_switch, c_switch, e1_switch, e2_switch, e3_switch, status_switch, turn_switch, adapt_button, help_button, rel_offset_switch
global scale_prop, x_prop, y_prop, z_prop, a_prop, b_prop, c_prop, e1_prop, e2_prop, e3_prop, status_prop, turn_prop

def first_state():
  '''
  First state that is called to start the script execution. See end of the script.
  '''
  #reload(ds_helpers)
  global init, app
  global scale_switch, x_switch, y_switch, z_switch, a_switch, b_switch, c_switch, e1_switch, e2_switch, e3_switch, status_switch, turn_switch, adapt_button, help_button, rel_offset_switch
  global scale_prop, x_prop, y_prop, z_prop, a_prop, b_prop, c_prop, e1_prop, e2_prop, e3_prop, status_prop, turn_prop

  # print(os.path.basename(__file__))
  init = True
  show_task_pane()
  init = False

  # display task pane in KUKA.Sim UI with all previously defined variables (prop1, prop2, etc.)
  executeInActionPanel()


def show_task_pane():
  """
  Set up task pane for UI in KUKA.Sim

  """
  global init, app
  global scale_switch, x_switch, y_switch, z_switch, a_switch, b_switch, c_switch, e1_switch, e2_switch, e3_switch, status_switch, turn_switch, adapt_button, help_button, rel_offset_switch
  global scale_prop, x_prop, y_prop, z_prop, a_prop, b_prop, c_prop, e1_prop, e2_prop, e3_prop, status_prop, turn_prop

  groupNr = 0  # groupNr is used to organize displayed elements in the task pane which is shown by executeInActionPanel()

  app = getApplication()
  scope = app.TeachContext.ActiveScope

  # get active robot from teach context
  rob = app.TeachContext.ActiveRobot
  if not rob:
    print (GLOBAL_ERROR + 'No program selected in teach page. Helper aborted!')
    return
  rob_controller = rob.findBehavioursByType(VC_ROBOTCONTROLLER)[0]

  sel_rob_info = createProperty(VC_STRING, 'Selected Robot')
  sel_rob_info.IsVisible = True
  ds_helpers.set_param_writable(sel_rob_info, False)
  sel_rob_info.Value = rob.Name # s.getSelection(VC_SELECTION_COMPONENT).Name
  sel_rob_info.Group = groupNr
  groupNr += 1

  scale_switch = createProperty(VC_BOOLEAN, 'Scaling')
  scale_switch.IsVisible = True
  ds_helpers.set_param_writable(scale_switch, True)
  scale_switch.OnChanged = show_updated_task_pane
  scale_switch.Group = groupNr
  groupNr += 1

  scale_prop = createProperty(VC_REAL, 'Scale_Factor')
  scale_prop.Value = 1.0
  if(scale_switch.Value):
    scale_prop.IsVisible = True
  else:
    scale_prop.IsVisible = False
  ds_helpers.set_param_writable(scale_prop, True)
  scale_prop.Group = groupNr
  groupNr += 1
  
  rel_offset_switch = createProperty(VC_BOOLEAN, 'Relative Offset')
  rel_offset_switch.IsVisible = True
  ds_helpers.set_param_writable(rel_offset_switch, True)
  rel_offset_switch.OnChanged = show_updated_task_pane
  rel_offset_switch.Group = groupNr
  groupNr += 1

  heading_1 = createProperty(VC_BUTTON, 'Select parameters to adapt')
  heading_1.IsVisible = True
  heading_1.WritableWhenConnected = False
  heading_1.WritableWhenDisconnected = False
  heading_1.Group = groupNr
  groupNr += 1

  x_switch = createProperty(VC_BOOLEAN, 'X')
  x_switch.IsVisible = True
  ds_helpers.set_param_writable(x_switch, True)
  x_switch.OnChanged = show_updated_task_pane
  x_switch.Group = groupNr
  groupNr += 1

  x_prop = createProperty(VC_REAL, 'X_Value')
  if(x_switch.Value):
    x_prop.IsVisible = True
  else:
    x_prop.IsVisible = False
  ds_helpers.set_param_writable(x_prop, True)
  x_prop.Group = groupNr
  groupNr += 1

  y_switch = createProperty(VC_BOOLEAN, 'Y')
  y_switch.IsVisible = True
  ds_helpers.set_param_writable(y_switch, True)
  y_switch.OnChanged = show_updated_task_pane
  y_switch.Group = groupNr
  groupNr += 1

  y_prop = createProperty(VC_REAL, 'Y_Value')
  if(y_switch.Value):
    y_prop.IsVisible = True
  else:
    y_prop.IsVisible = False
  ds_helpers.set_param_writable(y_prop, True)
  y_prop.Group = groupNr
  groupNr += 1

  z_switch = createProperty(VC_BOOLEAN, 'Z')
  z_switch.IsVisible = True
  ds_helpers.set_param_writable(z_switch, True)
  z_switch.OnChanged = show_updated_task_pane
  z_switch.Group = groupNr
  groupNr += 1

  z_prop = createProperty(VC_REAL, 'Z_Value')
  if(z_switch.Value):
    z_prop.IsVisible = True
  else:
    z_prop.IsVisible = False
  ds_helpers.set_param_writable(z_prop, True)
  z_prop.Group = groupNr
  groupNr += 1

  a_switch = createProperty(VC_BOOLEAN, 'A')
  a_switch.IsVisible = True
  ds_helpers.set_param_writable(a_switch, True)
  a_switch.OnChanged = show_updated_task_pane
  a_switch.Group = groupNr
  groupNr += 1

  a_prop = createProperty(VC_REAL, 'A_Value')
  if(a_switch.Value):
    a_prop.IsVisible = True
  else:
    a_prop.IsVisible = False
  ds_helpers.set_param_writable(a_prop, True)
  a_prop.Group = groupNr
  groupNr += 1

  b_switch = createProperty(VC_BOOLEAN, 'B')
  b_switch.IsVisible = True
  ds_helpers.set_param_writable(b_switch, True)
  b_switch.OnChanged = show_updated_task_pane
  b_switch.Group = groupNr
  groupNr += 1

  b_prop = createProperty(VC_REAL, 'B_Value')
  if(b_switch.Value):
    b_prop.IsVisible = True
  else:
    b_prop.IsVisible = False
  ds_helpers.set_param_writable(b_prop, True)
  b_prop.Group = groupNr
  groupNr += 1

  c_switch = createProperty(VC_BOOLEAN, 'C')
  c_switch.IsVisible = True
  ds_helpers.set_param_writable(c_switch, True)
  c_switch.OnChanged = show_updated_task_pane
  c_switch.Group = groupNr
  groupNr += 1

  c_prop = createProperty(VC_REAL, 'C_Value')
  if(c_switch.Value):
    c_prop.IsVisible = True
  else:
    c_prop.IsVisible = False
  ds_helpers.set_param_writable(c_prop, True)
  c_prop.Group = groupNr
  groupNr += 1

  status_switch = createProperty(VC_BOOLEAN, 'Status')
  status_switch.IsVisible = True
  ds_helpers.set_param_writable(status_switch, True)
  status_switch.OnChanged = show_updated_task_pane
  status_switch.Group = groupNr
  groupNr += 1
  status_prop = createProperty(VC_INTEGER, 'Status_Value')
  if(status_switch.Value):
    status_prop.IsVisible = True
  else:
    status_prop.IsVisible = False
  ds_helpers.set_param_writable(status_prop, True)
  status_prop.Group = groupNr
  groupNr += 1

  '''
  turn_switch = createProperty(VC_BOOLEAN, 'Turn')
  turn_switch.IsVisible = True
  ds_helpers.set_param_writable(turn_switch, True)
  turn_switch.OnChanged = show_updated_task_pane
  turn_switch.Group = groupNr
  groupNr += 1


  turn_prop = createProperty(VC_INTEGER, 'Turn_Value')
  if(turn_switch.Value):
    turn_prop.IsVisible = True
  else:
    turn_prop.IsVisible = False
  ds_helpers.set_param_writable(turn_prop, True)
  turn_prop.Group = groupNr
  groupNr += 1
  '''

  e1_switch = createProperty(VC_BOOLEAN, 'E1')
  e1_switch.IsVisible = True
  ds_helpers.set_param_writable(e1_switch, True)
  e1_switch.OnChanged = show_updated_task_pane
  e1_switch.Group = groupNr
  groupNr += 1

  e1_prop = createProperty(VC_REAL, 'E1_Value')
  if(e1_switch.Value):
    e1_prop.IsVisible = True
  else:
    e1_prop.IsVisible = False
  ds_helpers.set_param_writable(e1_prop, True)
  e1_prop.Group = groupNr
  groupNr += 1

  e2_switch = createProperty(VC_BOOLEAN, 'E2')
  e2_switch.IsVisible = True
  ds_helpers.set_param_writable(e2_switch, True)
  e2_switch.OnChanged = show_updated_task_pane
  e2_switch.Group = groupNr
  groupNr += 1

  e2_prop = createProperty(VC_REAL, 'E2_Value')
  if(e2_switch.Value):
    e2_prop.IsVisible = True
  else:
    e2_prop.IsVisible = False
  ds_helpers.set_param_writable(e2_prop, True)
  e2_prop.Group = groupNr
  groupNr += 1

  e3_switch = createProperty(VC_BOOLEAN, 'E3')
  e3_switch.IsVisible = True
  ds_helpers.set_param_writable(e3_switch, True)
  e3_switch.OnChanged = show_updated_task_pane
  e3_switch.Group = groupNr
  groupNr += 1

  e3_prop = createProperty(VC_REAL, 'E3_Value')
  if(e3_switch.Value):
    e3_prop.IsVisible = True
  else:
    e3_prop.IsVisible = False
  ds_helpers.set_param_writable(e3_prop, True)
  e3_prop.Group = groupNr
  groupNr += 1

  adapt_button = createProperty(VC_BUTTON, 'Adapt Statements')
  adapt_button.OnChanged = adapt_statements
  adapt_button.Group = groupNr
  groupNr += 1

  empty_line_1 = createProperty(VC_BUTTON, '----------------------------------------------------------------------------------------------------------------------------------------------------')
  empty_line_1.IsVisible = True
  empty_line_1.WritableWhenConnected = False
  empty_line_1.WritableWhenDisconnected = False
  empty_line_1.Group = groupNr
  groupNr += 1

  help_button = createProperty(VC_BUTTON, 'Open Help File')
  help_button.IsVisible = True
  help_button.OnChanged = open_help
  help_button.Group = groupNr
  groupNr += 1


def show_updated_task_pane(var):
  """
  Executes show_task_pane to update the UI.

  """

  global init, app
  global scale_switch, x_switch, y_switch, z_switch, a_switch, b_switch, c_switch, e1_switch, e2_switch, e3_switch, status_switch, turn_switch, adapt_button, help_button, rel_offset_switch
  global scale_prop, x_prop, y_prop, z_prop, a_prop, b_prop, c_prop, e1_prop, e2_prop, e3_prop, status_prop, turn_prop

  show_task_pane()


def calc_new_pos(parameter, offset, rel_offset, scale_factor):

  """
  Absolute or relative offset calculation of given parameter for the given offset. Returns updated parameter as Real value.

  :param float parameter: vector parameter that should be updated
  :param float offset: offset value to add on the vector parameter
  :param bool rel_offset: true - offset is calculated relative, false - offset is added as absolute value

  :returns float

  """
  par_updated = parameter

  if rel_offset:
    par_updated = (par_updated * scale_factor + offset)
  else:
    par_updated = offset * scale_factor

  return par_updated


def calc_pos_vector(pos_vector):
  """
  Calculates and returns translational vector with given offset values from action panel.

  :param Vector pos_vector: vector to apply offsets

  :returns Vector

  """

  position_vec = pos_vector

  # set position values if applicable
  if x_switch.Value:
    position_vec.X = calc_new_pos(position_vec.X, x_prop.Value, rel_offset_switch.Value, scale_prop.Value)
  if y_switch.Value:
    position_vec.Y = calc_new_pos(position_vec.Y, y_prop.Value, rel_offset_switch.Value, scale_prop.Value)
  if z_switch.Value:
    position_vec.Z = calc_new_pos(position_vec.Z, z_prop.Value, rel_offset_switch.Value, scale_prop.Value)

  return position_vec


def calc_rot_vector(rot_vector):
  """
  Calculates and returns rotational vector with given offset values from action panel.

  :param Vector pos_vector: vector to apply offsets

  :returns Vector

  """

  position_rot = rot_vector

  # set orientation values if applicable
  if a_switch.Value:
    # position_rot.Z = a_prop.Value
    position_rot.Z = calc_new_pos(position_rot.Z, a_prop.Value, rel_offset_switch.Value, scale_prop.Value)
  if b_switch.Value:
    # position_rot.Y = b_prop.Value
    position_rot.Y = calc_new_pos(position_rot.Y, b_prop.Value, rel_offset_switch.Value, scale_prop.Value)
  if c_switch.Value:
    # position_rot.X = c_prop.Value
    position_rot.X = calc_new_pos(position_rot.X, c_prop.Value, rel_offset_switch.Value, scale_prop.Value)

  return position_rot


def adapt_statements(var):
  """
  Adapts all selected statements according to chosen values.

  """

  sel_man = app.SelectionManager
  statements = sel_man.getSelection(VC_SELECTION_STATEMENT)

  for statement in statements:
    if 'spline' in statement.Name.lower():
      print(GLOBAL_ERROR + 'SPLINE block is not supported. Select all elements in the spline block individually!')
      return

    positions = statement.Positions # for CIRC you get more than one position from the statement
    for pos_index, position in enumerate(positions): # use enumerate to get the index of the current iteration
      position_mat = position.PositionInReference
      # ds_helpers.print_matrix_pos(position_mat) # debug print
      # print('---')# debug print

      # update position vector
      position_vec = position_mat.P
      position_mat.P = calc_pos_vector(position_vec)

      # update rotation vector
      position_rot = position_mat.WPR
      position_mat.WPR = calc_rot_vector(position_rot)

      # update position matrix in statement
      statement.Positions[pos_index].PositionInReference = position_mat

      # update status and turn
      if status_switch.Value:
        print(type(statement.Positions[pos_index]))
        new_status_val_binary_str = '{0:03b}'.format(status_prop.Value) # convert int value to binary with length of 3 bit
        statement.Positions[pos_index].Configuration = new_status_val_binary_str
        print(statement.Positions[pos_index].Configuration)


     ### ADAPTION OF TURN DOES NOT WORK FOR UNKNOWN REASONS
     ### Turn seems to be handled differently for robot model in KUKA.Sim (Visual Components) opposed to KSS
      '''
      if turn_switch.Value:
        new_turn_val = turn_prop.Value
        #print new_turn_val
        new_turn_val_binary_str = '{0:06b}'.format(new_turn_val) # convert int value to binary with length of 6 bit
        #print(new_turn_val_binary_str)
        new_turn_val_binary_list = [int(char) for char in new_turn_val_binary_str] # write bin string to int list through list comprehension
        #print(new_turn_val_binary_list)
        statement.Positions[pos_index].A1Turn = new_turn_val_binary_list[5]
        statement.Positions[pos_index].A2Turn = new_turn_val_binary_list[4]
        statement.Positions[pos_index].A3Turn = new_turn_val_binary_list[3]
        statement.Positions[pos_index].A4Turn = new_turn_val_binary_list[2]
        statement.Positions[pos_index].A5Turn = new_turn_val_binary_list[1]
        statement.Positions[pos_index].A6Turn = new_turn_val_binary_list[0]
        for prop in statement.Positions[pos_index].Properties:
          print prop.Name + ' =  ' + str(prop.Value) + ' (' + prop.Type + ')' 
      '''

      # set values for external axis if applicable
      if e1_switch.Value:
        position.E1 = calc_new_pos(position.E1, e1_prop.Value, rel_offset_switch.Value, scale_prop.Value)
      if e2_switch.Value:
        position.E2 = calc_new_pos(position.E2, e2_prop.Value, rel_offset_switch.Value, scale_prop.Value)
      if e3_switch.Value:
        position.E3 = calc_new_pos(position.E3, e3_prop.Value, rel_offset_switch.Value, scale_prop.Value)

      # ds_helpers.print_matrix_pos(position_mat) # debug print


  print ('Statements have been adapted according to chosen values')


def open_help(para):
  """
  Opens a help file for the add on.
  """
  file_path = ds_helpers.get_script_path(__file__)
  print('Open help file from ' + file_path)
  os.startfile(file_path + '\\Help_File_AdaptStatements.pdf')

# this executes the first_state method when the command is called from KUKA.Sim -> UI is opened
addState(first_state)
