"""
This helper originally has been created by an unknown KUKA.Sim user → KUDOS go out to the unknown user.
Original code has been embedded here, to fit to the DS-SIM Public Tools menu structure. User menu (task pane) has also been adapted.
"""
# Import for ds_sim_helpers library loading
import os
import sys
# import of stub file to enable auto completion for VC modules (C:\Python27\Lib\vcScript)
from vcScript import *
# Standard import for Python Addon
from vcCommand import *  # type: ignore
# standard kuka import
from kuka import * # type: ignore
# Loading regex lib
import re

import vcMatrix as mat

app = getApplication()
cmd = getCommand() # type: ignore

global baseNo, toolNo

default_task_pane_elements = []
base_task_pane_elements = []
tool_task_pane_elements = []

def first_state():
  default_task_pane()
  
  executeInActionPanel() # type: ignore


def default_task_pane():
  """
  default task pane creates all parameters that can be available or visible in the task pane.
  Visibility of certain parameters is set as necessary.
  """
  global baseNo, toolNo
  global default_task_pane_elements, base_task_pane_elements, tool_task_pane_elements
  
  description = cmd.createProperty(VC_STRING, 'Description')
  description.Value ="""
  
  1.  Select statements in program
      (statements without positions 
      are ignored)
  2.  Choose Tool or Base
  3.  Insert new Tool Nr. or Base Nr.
  4.  Click on execute
  
  """
  description.IsVisible = False
  
  change_base_button = cmd.createProperty(VC_BUTTON, 'Change Base')
  change_base_button.OnChanged = show_base_task_pane
  change_base_button.IsVisible = True
  default_task_pane_elements.append(change_base_button)
  
  change_tool_button = cmd.createProperty(VC_BUTTON, 'Change Tool')
  change_tool_button.OnChanged = show_tool_task_pane
  change_tool_button.IsVisible = True
  default_task_pane_elements.append(change_tool_button)
  
  baseNo = cmd.createProperty(VC_INTEGER, 'New Base', VC_PROPERTY_LIMIT)
  baseNo.MinValue = 0
  baseNo.MaxValue = 32
  baseNo.Value = 1
  baseNo.IsVisible = False
  base_task_pane_elements.append(baseNo)

  execute_base_change = cmd.createProperty(VC_BUTTON, 'Execute base change')
  execute_base_change.IsVisible = False
  execute_base_change.OnChanged = changeBase
  base_task_pane_elements.append(execute_base_change)
  
  toolNo = cmd.createProperty(VC_INTEGER, 'New Tool', VC_PROPERTY_LIMIT)
  toolNo.MinValue = 0
  toolNo.MaxValue = 16
  toolNo.Value = 1
  toolNo.IsVisible = False
  tool_task_pane_elements.append(toolNo)
  
  execute_tool_change = cmd.createProperty(VC_BUTTON, 'Execute tool change')
  execute_tool_change.IsVisible = False
  execute_tool_change.OnChanged = changeTool
  tool_task_pane_elements.append(execute_tool_change)
  
  # back to default screen
  go_back_button = cmd.createProperty(VC_BUTTON, 'Go Back')
  go_back_button.IsVisible = False
  go_back_button.OnChanged = show_default_task_pane
  tool_task_pane_elements.append(go_back_button)
  base_task_pane_elements.append(go_back_button)


def show_default_task_pane(var):
  global default_task_pane_elements, base_task_pane_elements, tool_task_pane_elements
  
  hide_base_task_pane()
  hide_tool_task_pane()
  for par in default_task_pane_elements:
    par.IsVisible = True


def hide_default_task_pane():
  global default_task_pane_elements
  
  for par in default_task_pane_elements:
    par.IsVisible = False


def show_base_task_pane(var):
  global default_task_pane_elements, base_task_pane_elements
  
  hide_default_task_pane()
  for par in base_task_pane_elements:
    par.IsVisible = True


def hide_base_task_pane():
  global default_task_pane_elements, base_task_pane_elements
  
  for par in base_task_pane_elements:
    par.IsVisible = False


def show_tool_task_pane(var):
  global default_task_pane_elements, tool_task_pane_elements
  
  hide_default_task_pane()
  for par in tool_task_pane_elements:
    par.IsVisible = True


def hide_tool_task_pane():
  global default_task_pane_elements, tool_task_pane_elements
  
  for par in tool_task_pane_elements:
    par.IsVisible = False


def changeBase(arg):
  global baseNo
  
  selection = app.SelectionManager.getSelection(VC_SELECTION_STATEMENT)
  
  matrix = mat.new()
  
  if len(selection) > 0:
    Positions = False
    for n in selection:
      if "vcPositionStatement" in str(type(n)):
        Positions = True
        break
    if Positions == True:
      for n in selection:
        if "vcPositionStatement" in str(type(n)):
          matrix = n.Positions[0].PositionInWorld
          n.BaseIndex = baseNo.Value
          n.Positions[0].PositionInWorld = matrix
    else:
      print("no positions selected")
  else:
    print("no positions selected")
    
  print('Change Base finished')


def changeTool(arg):
  global toolNo
  
  selection = app.SelectionManager.getSelection(VC_SELECTION_STATEMENT)
  
  matrix = mat.new()
  
  if len(selection) > 0:
    Positions = False
    for n in selection:
      if "vcPositionStatement" in str(type(n)):
        Positions = True
        break
    if Positions == True:
      for n in selection:
        if "vcPositionStatement" in str(type(n)):
          matrix = n.Positions[0].PositionInWorld
          try:
            tOld = n.Tool.PositionMatrix
          except:
            tOld = None
          n.ToolIndex = toolNo.Value
          try:
            tNew = n.Tool.PositionMatrix
          except:
            tNew = None
          if tOld and tNew:
            tOld.invert()
            tNew = tOld * tNew
          elif tOld:
            tOld.invert()
            tNew = tOld
          if tNew:
            matrix = matrix * tNew
            n.Positions[0].PositionInWorld = matrix
          
    else:
      print("no positions selected")
  else:
    print("no positions selected")

  print('Change Tool finished')


addState(first_state) # type: ignore


