# import of stub file to enable auto completion for VC modules (C:\Python27\Lib\vcScript)
from vcScript import *
# Standard import for Python Addon
from vcCommand import *  # type: ignore
from kuka import * # type: ignore

import vcHelpers.Selection as sel_help

import kuka_sim_extensions.Libs.DS_Library.ds_sim_helpers as ds_helpers

cmd = getCommand() # type: ignore
global is_init, comp, suffix, add_suffix_button, delete_suffix_button, set_marked_visible_button, set_marked_invisible_button, force_all_features_visible_button, force_all_features_invisible_button
is_init = False

def start_vis_assist():
  
  show_task_pane()


def show_task_pane():
    global is_init, comp, suffix, add_suffix_button, delete_suffix_button, set_marked_visible_button, set_marked_invisible_button, force_all_features_visible_button, force_all_features_invisible_button
    
    # Checks if exactly one comp is selected
    comp = ds_helpers.get_selected_comp()
    if not comp:
        nodes = sel_help.getSelectedNodes()
        if len(nodes) <= 0:
          print('No component or nodes selected. Please select a component or at least one node.')
          return
    else:
      nodes = sel_help.getGivenComponentsNodes(comp)
    
    # with property VC_STRING Description you can write a description text in the task pane
    description = cmd.createProperty(VC_STRING, 'Description')
    description.IsVisible = False
    description.Value = """
    Add suffix to all children
    1. Select a component or node of a component
    2. Enter a suffix in the suffix field
    3. Click on "Add_Suffix" to add suffix to all invisible child features
    
    Delete suffix of all children
    1. Select a component or node of a component
    2. Check if suffix field fits to suffix that should be deleted
    3. Click on "Delete_Suffix" to delete suffixes from child features
    
    Set_Marked_Visible
    - set all features marked with the suffix visible
    
    Set_Marked_Invisible
    - set all features marked with the suffix invisible
    """
    description.IsVisible = False
    
    suffix = cmd.createProperty(VC_STRING, 'Suffix')
    if not is_init:
      suffix.Value = '_INVIS'
    
    add_suffix_button = cmd.createProperty(VC_BUTTON, 'Add_Suffix')
    add_suffix_button.OnChanged = add_suffix
    
    delete_suffix_button = cmd.createProperty(VC_BUTTON, 'Delete_Suffix')
    delete_suffix_button.OnChanged = delete_suffix
    
    set_marked_visible_button = cmd.createProperty(VC_BUTTON, 'Set_Marked_Visible')
    set_marked_visible_button.OnChanged = set_marked_visible
    
    set_marked_invisible_button = cmd.createProperty(VC_BUTTON, 'Set_Marked_Invisible')
    set_marked_invisible_button.OnChanged = set_marked_invisible
    
    force_all_features_visible_button = cmd.createProperty(VC_BUTTON, 'Force_All_Features_Visible')
    force_all_features_visible_button.OnChanged = force_all_features_visible
    
    force_all_features_invisible_button = cmd.createProperty(VC_BUTTON, 'Force_All_Features_Invisible')
    force_all_features_invisible_button.OnChanged = force_all_features_invisible
    
    # Executes the Command in the ActionPanel => Action Panel Opens with all the Attributes of the Command
    executeInActionPanel() # type: ignore


def add_suffix(var):
  global suffix
  nodes = update_selection()
  if nodes:
    node_crawler_tagging(nodes[0], suffix.Value, True)


def delete_suffix(var):
  global suffix
  nodes = update_selection()
  if nodes:
    node_crawler_tagging(nodes[0], suffix.Value, False)


def set_marked_visible(var):
  global suffix
  nodes = update_selection()
  if nodes:
    node_crawler_visibility(nodes[0], suffix.Value, True, False)
    update_sim_env(nodes)


def set_marked_invisible(var):
  global suffix
  nodes = update_selection()
  if nodes:
    node_crawler_visibility(nodes[0], suffix.Value, False, False)
    update_sim_env(nodes)


def force_all_features_visible(var):
  global suffix
  nodes = update_selection()
  if nodes:
    node_crawler_visibility(nodes[0], suffix.Value, True, True)
    update_sim_env(nodes)


def force_all_features_invisible(var):
  global suffix
  nodes = update_selection()
  if nodes:
    node_crawler_visibility(nodes[0], suffix.Value, False, True)
    update_sim_env(nodes)


def update_sim_env(nodes):
  nodes[0].rebuild()
  getApplication().render()


def update_selection():
  comp = ds_helpers.get_selected_comp()
  if not comp:
      nodes = sel_help.getSelectedNodes()
      if len(nodes) <= 0:
        print('No component or nodes selected. Please select a component or at least one node.')
        return None
  else:
    nodes = sel_help.getGivenComponentsNodes(comp)
  
  return nodes


def node_crawler_visibility(node, suffix, visibility, force):
  """ Iterates through node structure (e.g. links) of a component.
  """
  # print(node.Name) # debug print
  feature_crawler_visibility(node.RootFeature, suffix, visibility, force)
  nodeList = node.Children
  for node in nodeList:
    node_crawler_visibility(node, suffix, visibility, force)

def node_crawler_tagging(node, suffix, tagging):
  """ Iterates through node structure (e.g. links) of a component.
  """
  # print(node.Name) # debug print
  feature_crawler_tagging(node.RootFeature, suffix, tagging)
  nodeList = node.Children
  for node in nodeList:
    node_crawler_tagging(node, suffix, tagging)


def feature_crawler_visibility(feature, suffix, visibility, force):
  """ find all features with matching suffix 
  """
  if feature.Name != 'Root':
    # print(feature.Name) # debug print
    if force == True:
      feature.Visible = visibility
    else:  
      if suffix in feature.Name:
        feature.Visible = visibility
        #print(feature.Name + 'visibility changed')
  
  child_feat_list = feature.Children
  if len(child_feat_list) > 0:
    for feature in child_feat_list:
      # print(feature.Name) # debug print
      feature_crawler_visibility(feature, suffix, visibility, force)  


def feature_crawler_tagging(feature, suffix, tagging):
  """ find all features with matching suffix 
  """
  if feature.Name != 'Root':
    if tagging:
      if not feature.Visible and suffix not in feature.Name:
        feature.Name = feature.Name + suffix
    else:
      feature.Name = feature.Name.replace(suffix, '')
  
  child_feat_list = feature.Children
  if len(child_feat_list) > 0:
    for feature in child_feat_list:
      # print(feature.Name) # debug print
      feature_crawler_tagging(feature, suffix, tagging)  


def feature_manipulator(feature, suffix, visibility, force_visible=False):
  """ Visibility of given feature and all of its child features is set to True 
  """
  featureList = feature.Children
  feature.Visible = True
  for feature in featureList:
    print(feature.Name)
    if len(featureList) > 0:
      feature.Visible = True
      feature_manipulator(feature, suffix, visibility, force_visible)
    else:
      if force_visible:
        feature.Visible = True
      else:
        if suffix in feature.Name:
          feature.Visible = visibility

addState(start_vis_assist) # type: ignore