"""
This addon mirrors selected statements for (S)PTP, (S)LIN, (S)CIRC, SPLINE-Blocks.

20.09.2022 - V 0.0 - Prototype
29.09.2022 - V 1.0 - added CIRC support, global params and custom mode for user
05.10.2022 - V 1.1 - added support for SPLINE blocks. Adapation of code for statements that worked with V 1.0 was also necessary (only for Parent_base)
24.04.2023 - V 1.2 - added help button
25.04.2023 - V 1.3 - fixed bug regarding SPLINE -> path vel is now updated for mirrored SPLINE
"""

import os
import sys
from vcCommand import *
from vcHelpers.Selection import *
from kuka import *
import vcMatrix
import kuka_sim_extensions.Libs.DS_Library.ds_sim_helpers as ds_helpers

GLOBAL_ERROR = "MIRROR WIZARD ERROR -"

def first_state():
  #reload(ds_helpers)
  global init, app, rob, rob_controller, chosen_mirror_base, mirror_plane, mirror_pos_button, show_params, mirror_params_mode, help_button
  global x_mod, y_mod, z_mod, a_mod, b_mod, c_mod, a_offset, b_offset, c_offset
  global x_mod_par, y_mod_par, z_mod_par, a_mod_par, b_mod_par, c_mod_par, a_offset_par, b_offset_par, c_offset_par
  
  # print(os.path.basename(__file__))
  init = True
  show_task_pane()
  init = False

  # display task pane in KUKA.Sim UI with all previously defined variables (prop1, prop2, etc.)
  executeInActionPanel()

def show_task_pane():
  """
  Set up task pane for UI in KUKA.Sim

  """
  global init, app, rob, rob_controller, chosen_mirror_base, mirror_plane, mirror_pos_button, show_params, mirror_params_mode, help_button
  global x_mod, y_mod, z_mod, a_mod, b_mod, c_mod, a_offset, b_offset, c_offset
  global x_mod_par, y_mod_par, z_mod_par, a_mod_par, b_mod_par, c_mod_par, a_offset_par, b_offset_par, c_offset_par

  groupNr = 0  # groupNr is used to organize displayed elements in the task pane which is shown by executeInActionPanel()

  app = getApplication()
  scope = app.TeachContext.ActiveScope

  # get active robot from teach context
  rob = app.TeachContext.ActiveRobot
  if not rob:
    print (GLOBAL_ERROR + 'No program selected in teach page. Wizard aborted!')
    return
  rob_controller = rob.findBehavioursByType(VC_ROBOTCONTROLLER)[0]
  rob_bases = rob_controller.Bases
  rob_tools = rob_controller.Tools

  sel_rob_info = createProperty(VC_STRING, 'Selected Robot')
  sel_rob_info.IsVisible = True
  set_param_writable(sel_rob_info, False)
  sel_rob_info.Value = rob.Name # s.getSelection(VC_SELECTION_COMPONENT).Name
  sel_rob_info.Group = groupNr
  groupNr += 1

  chosen_mirror_base = createProperty(VC_STRING, 'Mirror Base', VC_PROPERTY_STEP)
  names_rob_bases = [base.Name for base in rob_bases] # one liner to write all base names to a new list
  names_rob_bases.insert(0, 'Parent_Base')
  chosen_mirror_base.StepValues = names_rob_bases
  if init:
    chosen_mirror_base.Value = names_rob_bases[0]
  chosen_mirror_base.Group = groupNr
  groupNr += 1

  mirror_plane = createProperty(VC_STRING, 'Mirror Plane', VC_PROPERTY_STEP)
  mirror_plane_options = ['XZ', 'YZ']
  mirror_plane.StepValues = mirror_plane_options
  if init:
    mirror_plane.Value = mirror_plane_options[0]
  mirror_plane.OnChanged = show_updated_task_pane
  mirror_plane.Group = groupNr
  groupNr += 1

  mirror_params_mode = createProperty(VC_STRING, 'Mirror Parameters', VC_PROPERTY_STEP)
  mirror_params_options = ['Default', 'Custom']
  mirror_params_mode.StepValues = mirror_params_options
  if init:
    mirror_params_mode.Value = mirror_params_options[0]
  mirror_params_mode.OnChanged = show_updated_task_pane
  mirror_params_mode.Group = groupNr
  groupNr += 1

  show_params = createProperty(VC_BOOLEAN, 'Show Parameters')
  show_params.IsVisible = True
  if init:
    show_params.Value = False
  show_params.OnChanged = show_updated_task_pane
  show_params.Group = groupNr
  groupNr += 1

  x_mod_par = createProperty(VC_REAL, 'X_mod')
  if show_params.Value:
    x_mod_par.IsVisible = True
  else:
    x_mod_par.IsVisible = False
  if mirror_params_mode.Value == 'Custom':
    set_param_writable(x_mod_par, True)
  else:
    set_param_writable(x_mod_par, False)
  x_mod_par.Group = groupNr
  groupNr += 1

  y_mod_par = createProperty(VC_REAL, 'Y_mod')
  if show_params.Value:
    y_mod_par.IsVisible = True
  else:
    y_mod_par.IsVisible = False
  if mirror_params_mode.Value == 'Custom':
    set_param_writable(y_mod_par, True)
  else:
    set_param_writable(y_mod_par, False)
  y_mod_par.Group = groupNr
  groupNr += 1

  z_mod_par = createProperty(VC_REAL, 'Z_mod')
  if show_params.Value:
    z_mod_par.IsVisible = True
  else:
    z_mod_par.IsVisible = False
  if mirror_params_mode.Value == 'Custom':
    set_param_writable(z_mod_par, True)
  else:
    set_param_writable(z_mod_par, False)
  z_mod_par.Group = groupNr
  groupNr += 1

  a_mod_par = createProperty(VC_REAL, 'A_mod')
  if show_params.Value:
    a_mod_par.IsVisible = True
  else:
    a_mod_par.IsVisible = False
  if mirror_params_mode.Value == 'Custom':
    set_param_writable(a_mod_par, True)
  else:
    set_param_writable(a_mod_par, False)
  a_mod_par.Group = groupNr
  groupNr += 1

  b_mod_par = createProperty(VC_REAL, 'B_mod')
  if show_params.Value:
    b_mod_par.IsVisible = True
  else:
    b_mod_par.IsVisible = False
  if mirror_params_mode.Value == 'Custom':
    set_param_writable(b_mod_par, True)
  else:
    set_param_writable(b_mod_par, False)
  b_mod_par.Group = groupNr
  groupNr += 1

  c_mod_par = createProperty(VC_REAL, 'C_mod')
  if show_params.Value:
    c_mod_par.IsVisible = True
  else:
    c_mod_par.IsVisible = False
  if mirror_params_mode.Value == 'Custom':
    set_param_writable(c_mod_par, True)
  else:
    set_param_writable(c_mod_par, False)
  c_mod_par.Group = groupNr
  groupNr += 1

  a_offset_par = createProperty(VC_REAL, 'A_offset')
  if show_params.Value:
    a_offset_par.IsVisible = True
  else:
    a_offset_par.IsVisible = False
  if mirror_params_mode.Value == 'Custom':
    set_param_writable(a_offset_par, True)
  else:
    set_param_writable(a_offset_par, False) 
  a_offset_par.Group = groupNr
  groupNr += 1

  b_offset_par = createProperty(VC_REAL, 'B_offset')
  if show_params.Value:
    b_offset_par.IsVisible = True
  else:
    b_offset_par.IsVisible = False
  if mirror_params_mode.Value == 'Custom':
    set_param_writable(b_offset_par, True)
  else:
    set_param_writable(b_offset_par, False)
  b_offset_par.Group = groupNr
  groupNr += 1

  c_offset_par = createProperty(VC_REAL, 'C_offset')
  if show_params.Value:
    c_offset_par.IsVisible = True
  else:
    c_offset_par.IsVisible = False
  if mirror_params_mode.Value == 'Custom':
    set_param_writable(c_offset_par, True)
  else:
    set_param_writable(c_offset_par, False)
  c_offset_par.Group = groupNr
  groupNr += 1

  mirror_pos_button = createProperty(VC_BUTTON, 'Mirror Positions')
  mirror_pos_button.IsVisible = True
  mirror_pos_button.OnChanged = mirror_positions
  mirror_pos_button.Group = groupNr
  groupNr += 1

  empty_line = createProperty(VC_BUTTON, '----------------------------------------------------------------------------------------------------------------------------------------------------')
  empty_line.IsVisible = True
  empty_line.WritableWhenConnected = False
  empty_line.WritableWhenDisconnected = False
  empty_line.Group = groupNr
  groupNr += 1

  help_button = createProperty(VC_BUTTON, 'Open Help File')
  help_button.IsVisible = True
  help_button.OnChanged = open_help
  help_button.Group = groupNr
  groupNr += 1

  # set all global parameters (task pane, modificators and offsets for mirror calculation)
  set_params()

  init = False

def set_param_writable(param, boolean_value=False):
  """
  Sets all wirtable options of a paramter at once.

  """
  param.WritableWhenConnected = boolean_value
  param.WritableWhenDisconnected = boolean_value
  param.WritableWhenSimulating = boolean_value

def show_updated_task_pane(self):
  """
  Executes show_task_pane to update the UI.

  """
  global init, app, rob, rob_controller, chosen_mirror_base, mirror_plane, mirror_pos_button, show_params, mirror_params_mode
  global x_mod, y_mod, z_mod, a_mod, b_mod, c_mod, a_offset, b_offset, c_offset
  global x_mod_par, y_mod_par, z_mod_par, a_mod_par, b_mod_par, c_mod_par, a_offset_par, b_offset_par, c_offset_par

  show_task_pane()


def set_params():
  """
  This function sets and updates all global parameters, modificators and offsets that are relevant for mirroring.
  The values are determined from the chosen task pane user parameters e.g. mirror pane.

  """
  global init, app, rob, rob_controller, chosen_mirror_base, mirror_plane, mirror_pos_button, show_params, mirror_params_mode
  global x_mod, y_mod, z_mod, a_mod, b_mod, c_mod, a_offset, b_offset, c_offset
  global x_mod_par, y_mod_par, z_mod_par, a_mod_par, b_mod_par, c_mod_par, a_offset_par, b_offset_par, c_offset_par

  x_mod = y_mod = z_mod = a_mod = b_mod = c_mod = 1.0
  a_offset = b_offset = c_offset = 0.0

  # set default values for params depending on chosen mirror plane
  if mirror_plane.Value == 'XZ':
    y_mod = -1.0
    c_mod = -1.0
    a_mod = -1.0
  elif mirror_plane.Value == 'YZ':
    x_mod = -1.0
    c_mod = -1.0
    a_mod = -1.0
    a_offset = 180.0
  else:
    print(GLOBAL_ERROR + ': unsupported mirror plane')

  # write all values to parameters in the task pane as help for user
  x_mod_par.Value = x_mod
  y_mod_par.Value = y_mod
  z_mod_par.Value = z_mod
  a_mod_par.Value = a_mod
  b_mod_par.Value = b_mod
  c_mod_par.Value = c_mod
  a_offset_par.Value = a_offset
  b_offset_par.Value = b_offset
  c_offset_par.Value = c_offset

  if mirror_params_mode.Value == 'Default':
    # we just use the default parameters so nothing to do here
    pass
  elif mirror_params_mode.Value == 'Custom':
    # reread the parameters from the task pane, as the user might have changed them
    x_mod, y_mod, z_mod, a_mod, b_mod, c_mod = [x_mod_par.Value, y_mod_par.Value, z_mod_par.Value, a_mod_par.Value, b_mod_par.Value, c_mod_par.Value]
    a_offset, b_offset, c_offset = [a_offset_par.Value, b_offset_par.Value, c_offset_par.Value]
  else:
    print(GLOBAL_ERROR + ' unsupported Mirror Parameters')


def mirror_positions(para):
  """
  Checks if statements have been selected, adds a new fold to the program and triggers the next method to calculate
  and mirror the positions from the selcted statemens.
  """

  global app, rob, chosen_mirror_base, mirror_plane
  sel_man = app.SelectionManager

  statements = sel_man.getSelection(VC_SELECTION_STATEMENT)
  if len(statements) == 0:
    print(GLOBAL_ERROR + ' No position statements selected, aborting wizard!')
    return

  print('Number of selected statements: ' + str(len(statements)))

  # add new fold for mirrored positions
  active_routine = app.TeachContext.ActiveRoutine
  scope = addFoldStatement(active_routine, 'Mirrored_Positions')

  # mirror positions
  calc_and_add_mirror_positions(scope, statements, chosen_mirror_base)

  print('Positions mirrored')


def calc_and_add_mirror_positions(scope, statements, chosen_mirror_base):
  """
  Calculates the mirrored positions for a list of statements in reference to parent base for a chosen mirror plane (global) and adds them to the given scope.

  :param Scope scope: scope where the mirrored statement will be added
  :param List statements: list of statements to mirror
  :param String chosen_mirror_base: string that represents the chosen mirror base from the user

  """
  global app, rob
  global x_mod, y_mod, z_mod, a_mod, b_mod, c_mod, a_offset, b_offset, c_offset

  if not chosen_mirror_base.Value == 'Parent_Base':
    mirror_base_nr = ds_helpers.get_rob_base_index(rob, chosen_mirror_base.Value)
    mirror_base = rob_controller.Bases[mirror_base_nr]

  mirror_positions_list = []
  for statement in statements:
    motion_type = statement.InlineFormIdentifier
    point_names = []
    ptp_vel = path_vel = 0.0
    base_index = None
    tool_index = None
    if 'ptp' in motion_type.lower():
      ptp_vel = statement.PtpVelocity
      point_names.append(statement.PointName)
    elif 'spline' in motion_type.lower():
      # print('SPLINE found')
      path_vel = statement.PathVelocity
      point_names.append(statement.SplineBlockPointName)
    elif 'spl' in motion_type.lower():
      path_vel = statement.PathVelocity
      point_names.append(statement.PointName)
    elif 'lin' in motion_type.lower():
      path_vel = statement.PathVelocity
      point_names.append(statement.PointName)
    elif 'circ' in motion_type.lower():
      path_vel = statement.PathVelocity
      point_names.append(statement.HelpPointName)
      point_names.append(statement.PointName)

    if ('spline' in motion_type.lower()):
      blend = 'None'
      base_index = statement.BaseIndex
      tool_index = statement.ToolIndex
      point_names.append(statement.SplineBlockPointName)
    elif ('spl' in motion_type.lower()):
      blend = 'None'
      # spl does not have a base index parameter, but we need this parameter later when calculating mirrored positions, so we get it from the parent statement
      base_index = statement.ParentScope.ParentStatement.BaseIndex
      tool_index = statement.ParentScope.ParentStatement.ToolIndex
    else:
      blend = statement.BlendingType
      base_index = statement.BaseIndex
      tool_index = statement.ToolIndex

    if base_index == 0:
      print(GLOBAL_ERROR + ' Base Null (BASE_DATA[0]) is not supported (see statement %s, line(s) %s). Mirror result may not be plausible!' %(statement.Name, statement.LineSpan))

    apo_dist = statement.ApproximationDistance

    pos_points = []

    if not ('spline' in motion_type.lower()):
      pos_points = statement.Positions

    mirror_positions = []
    for mirror_position in pos_points:

      if chosen_mirror_base.Value == 'Parent_Base':
        mirror_pos = mirror_position.PositionInReference
      else:
        # calculate position to be mirrored in reference to the chosen mirror base
        mirror_base_inv = rob.WorldPositionMatrix * rob_controller.WorldTransformMatrix * mirror_base.PositionMatrix
        mirror_base_inv.invert()
        mirror_pos = mirror_base_inv * mirror_position.PositionInWorld

      # ds_helpers.print_matrix_pos(mirror_pos)
      mirror_pos_vec = mirror_pos.P
      mirror_pos_vec.X = mirror_pos_vec.X * x_mod
      mirror_pos_vec.Y = mirror_pos_vec.Y * y_mod
      mirror_pos_vec.Z = mirror_pos_vec.Z * z_mod
      mirror_pos.P = mirror_pos_vec

      mirror_pos_rot = mirror_pos.WPR
      mirror_pos_rot.X = mirror_pos_rot.X * c_mod + c_offset
      mirror_pos_rot.Y = mirror_pos_rot.Y * b_mod + b_offset
      mirror_pos_rot.Z = mirror_pos_rot.Z * a_mod + a_offset
      mirror_pos.WPR = mirror_pos_rot

      mirror_positions.append(mirror_pos)

    statement_scope = statement.Scope
    mirror_positions_list.append((motion_type, ptp_vel, path_vel, point_names, base_index, tool_index, blend, apo_dist, mirror_positions, statement_scope))

  add_statements(scope, mirror_positions_list, chosen_mirror_base)


def add_statements(scope, position_tuples, chosen_mirror_base=None):
  """
  This function adds statements to the given scope.

  :param Scope scope: scope where the statements will be added
  :param Tuple List: position_tuples: list of tuples, tuples carry all infos that are required to create motion statements
                     (motion_type, ptp_vel, path_vel, point_names, base_index, tool_index, blend, apo_dist, mirror_positions, statement_scope)
  :param String chosen_mirror_base: string that represents the chosen mirror base from the user

  """
  # add statements to scope (e.g. a fold)
  for pos_tuple in position_tuples:
    motion_type, ptp_vel, path_vel, point_names, base_index, tool_index, blend, apo_dist, mirror_positions, statement_scope = pos_tuple
    rob_bases = rob_controller.Bases

    if 'ptp' in motion_type.lower():
      parameters = 'BaseIndex=%i, ToolIndex=%i, BlendingType=%s, ApproximationDistance=%f, PointName=%s, PtpVelocity=%f' % (base_index, tool_index, blend, apo_dist, point_names[0], ptp_vel)
    elif 'lin' in motion_type.lower():
      parameters = 'BaseIndex=%i, ToolIndex=%i, BlendingType=%s, ApproximationDistance=%f, PointName=%s, PathVelocity=%f' % (base_index, tool_index, blend, apo_dist, point_names[0], path_vel)
    elif 'circ' in motion_type.lower():
      parameters = 'BaseIndex=%i, ToolIndex=%i, BlendingType=%s, ApproximationDistance=%f, PathVelocity=%f' % (base_index, tool_index, blend, apo_dist, path_vel) # CIRC point names are added later on
    elif 'spl' in motion_type.lower() and 'spline' not in motion_type.lower():
      parameters = 'ApproximationDistance=%f, PathVelocity=%f' % (apo_dist, path_vel)
    elif 'spline' in motion_type.lower():
      # no parameters are set for the creation of a spline block, as addKukaSplineBlock(scope, index=-1) does not offer to hand in parameters
      pass
    else:
      parameters = 'BaseIndex=%i, ToolIndex=%i, BlendingType=%s, ApproximationDistance=%f' % (base_index, tool_index, blend, apo_dist)

    mirror_positions_world = []
    for mirror_position in mirror_positions:
      base = rob_bases[base_index - 1] # base index 0 -> Base_DATA[0] for KUKA robots, base index 0 -> Base_DATA[1] for rob controller in KUKA.Sim
      if chosen_mirror_base.Value == 'Parent_Base':
        print('Mirroring positions to parent')
        if base_index != 0:
          mirror_pos_world = rob.WorldPositionMatrix * rob_controller.WorldTransformMatrix * base.PositionMatrix * mirror_position
        else:
          mirror_pos_world = rob.WorldPositionMatrix * rob_controller.WorldTransformMatrix * mirror_position
      else:
        print('Mirroring positions to mirror base')
        mirror_base = rob_bases[ds_helpers.get_rob_base_index(rob, chosen_mirror_base.Value)]
        mirror_pos_world = rob.WorldPositionMatrix * rob_controller.WorldTransformMatrix * mirror_base.PositionMatrix * mirror_position

      mirror_positions_world.append(mirror_pos_world)

    if 'circ' in motion_type.lower():
      help_mirror_pos_name = point_names[0] + "_"
      mirror_pos_name = point_names[1] + "_"
      # addExpertCirc(fold, motion_kind, pointExpression=None, helpPointExpression=None, index=-1) # this does not work
      new_statement = addKukaStatement(scope, motion_type, -1, parameters)
      new_statement.Positions[0].PositionInWorld = mirror_positions_world[0] # help point is on index 0 in positions of statement
      new_statement.Positions[0].Name = help_mirror_pos_name
      new_statement.Positions[1].PositionInWorld = mirror_positions_world[1] # target point is on index 1 in positions of statement
      new_statement.Positions[1].Name = mirror_pos_name
    elif 'spline' in motion_type.lower():
      # create new spline block
      mirror_spline = addKukaSplineBlock(scope)
      # set all parameters on by one as no parameter string is accepted by addKukaSplineBlock
      mirror_spline.SplineBlockPointName = point_names[0] + "_"
      mirror_spline.BaseIndex = base_index
      mirror_spline.ToolIndex = tool_index
      mirror_spline.PathVelocity = path_vel
      spline_statements = statement_scope.Statements
      mirror_spline_scope = mirror_spline.Scope
      # for a SPLINE we need to mirror all child statements of the SPLINE and add them in the new mirrored SPLINE
      # therefore we recursively call the calc_and_add_mirror_positions function and hand over all spline statements
      calc_and_add_mirror_positions(mirror_spline_scope, spline_statements, chosen_mirror_base)
    else:
      mirror_pos_name = point_names[0] + "_"
      new_statement = addKukaStatement(scope, motion_type, -1, parameters)
      updatePositionStatement(new_statement, mirror_positions_world[0], mirror_pos_name)
      # temp_pos = vcMatrix.new()#mirror_positions_world[0]
      # temp_pos.P = mirror_positions_world[0].P
      # temp_pos.WPR = mirror_positions_world[0].WPR
      # temp_pos.translateRel(0, 0, 200)
      # setPosition(new_statement, temp_pos)
      # setPosition(new_statement,  mirror_positions_world[0])
      # print('added')



def open_help(para):
  """
  Opens a help file for the add on.
  """
  file_path = ds_helpers.get_script_path(__file__)
  print('Open help file from ' + file_path)
  os.startfile(file_path + '\\Help_File_MirrorWizard.pdf')

# this executes the first_state method when the command is called from KUKA.Sim -> UI is opened
addState(first_state)
