"""
This wizard adds functionality to a selected component to show/hide this component with a Boolean signal.
"""
# Imported for ds_sim_helpers library loading
import os
import sys
# import of stub file to enable auto completion for VC modules (C:\Python27\Lib\vcScript)
from vcScript import *
# Standard import for Python Addon
from vcCommand import *  # type: ignore
# standard kuka import
from kuka import * # type: ignore
# Loading regex lib
import re
#importing ds_sim_helpers Library for getting selected component
import kuka_sim_extensions.Libs.DS_Library.ds_sim_helpers as ds_helpers

script_code ="""
from vcScript import *

comp = getComponent()
show_hide_switch = comp.getProperty('Show_Hide_Switch')

def OnSignal( signal ):

  if signal.Name == 'Show_Hide':
    show_hide_switch.Value = signal.Value

"""

app = getApplication()
sim = getSimulation()


def first_state():
  
  comp = ds_helpers.get_selected_comp()
  
  if not comp:
    return
  
  if not comp.getProperty('Show_Hide_Switch'):
    comp.createProperty(VC_BOOLEAN, 'Show_Hide_Switch')
  vis_switch = comp.getProperty('Show_Hide_Switch')
  vis_switch.Value = True

  if not comp.findBehaviour('Show_Hide_Script'):
    comp.createBehaviour(VC_PYTHONSCRIPT, 'Show_Hide_Script')
  show_hide_script = comp.findBehaviour('Show_Hide_Script')
  show_hide_script.Script = script_code
  
  
  if not comp.findNode('Show_Hide_Link'):
    comp.createNode(VC_NODE, 'Show_Hide_Link')
  show_hide_Link = comp.findNode('Show_Hide_Link')
  show_hide_Link.JointType = VC_JOINTTYPE_CUSTOM
  show_hide_Link.JointExpression = 'Show_Hide_Switch?Tx(0):Sx(0).Sy(0).Sz(0)'
  
  if not comp.findBehaviour('Show_Hide'):
    comp.createBehaviour(VC_BOOLEANSIGNAL, 'Show_Hide')
  show_hide_input = comp.findBehaviour('Show_Hide')
  show_hide_input.Connections = [show_hide_script]
  
  comp_root = comp.getFeature('Root')
  features = comp_root.Children
  show_hide_Link_root = show_hide_Link.getFeature('Root')
  for feat in features:
    show_hide_Link_root.attach(feat, True)
  
  comp.rebuild()
  
  print('Show/Hide wizard finished')

addState(first_state) # type: ignore
