import os
from vcCommand import * # type: ignore <- to ignore missing import in Pylance
from vcHelpers.Selection import *


def test_print(print_message):
  print(print_message)

def print_matrix_pos(mat):
  """
  This function prints a given position matrix with readable layout and annotations for X, Y, Z and A, B, C

  :param vcMatrix mat: The matrix to print

  """
  # print 'X = %f' % mat.P.X, '\tY = %f' % mat.P.Y, '\tZ = %f' % mat.P.Z
  print('X = %f' % (mat.P.X) + '\tY = %f' % (mat.P.Y) + '\tZ = %f' % (mat.P.Z))
  print('A = %f' % (mat.getEuler().Z) + '\tB = %f' % (mat.getEuler().Y) + '\tC = %f' % (mat.getEuler().X))

def get_selected_comp():
  """
  Checks if only one component is selected. If so, this component is returned

  """
  comps = getSelectedComponents()

  if len(comps) < 1:
    print("No component selected, aborting command!")
    return None
  if len(comps) > 1:
    print("More than one component selected, aborting command!")
    return None
 
  return comps[0]


def get_selected_robot():
  """
  Checks if the selected component is a robot and returns the robot.

  """
  rob = get_selected_comp()
  if rob is None:
    return
  if rob.Category != 'Robots':
    print('Selected component is not a robot, aborting command!')
    return

  return rob


def get_rob_base_index(rob, base_name):
  """
  Returns the index for the given base name of the given robot

  :param Robot rob: Robot that carries the base
  :param String base_name: Name of the base

  """
  rob_controller = rob.findBehavioursByType(VC_ROBOTCONTROLLER)[0]
  rob_bases = rob_controller.Bases
  rob_base_names = []
  for rob_base in rob_bases:
    rob_base_names.append(rob_base.Name)

  return rob_base_names.index(base_name)

def get_tool_base_index(rob, tool_name):
  """
  Returns the index for the given tool name of the given robot

  :param Robot rob: Robot that carries the base
  :param String tool_name: Name of the tool

  """
  rob_controller = rob.findBehavioursByType(VC_ROBOTCONTROLLER)[0]
  rob_tools = rob_controller.Tools
  rob_tool_names = []
  for rob_base in rob_tools:
    rob_tool_names.append(rob_base.Name)

  return rob_tool_names.index(tool_name)

def get_script_path(file):
  """
  Get script path and cut the 8 leading chars
    file:///C:\Users\sturma\OneDrive - KUKA AG\Dokumente\KUKA\KUKA.Sim 4.1\My Commands\SaveLoadSignalConnections
    ->
    C:\Users\sturma\OneDrive - KUKA AG\Dokumente\KUKA\KUKA.Sim 4.1\My Commands\SaveLoadSignalConnections
  
  """

  script_path = os.path.dirname(file)
  script_path_list = list(script_path)
  script_path_list[0:8] = ""
  script_path = "".join(script_path_list)

  return script_path

def set_param_writable(param, boolean_value=False):
  """
  Sets all wirtable options of a paramter at once.

  """
  param.WritableWhenConnected = boolean_value
  param.WritableWhenDisconnected = boolean_value
  param.WritableWhenSimulating = boolean_value